#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"SimpleImage.h"
#include"BMPIO.h"
#define PI 3.141592653589793238462643383279502884197

/* DoG画像を生成 */
Image* makeDoGImage(int radii,double sig,double K){
  int gsx = 2*radii+1;
  int gsy = gsx;
  double sig2 = 2.0*sig*sig;
  double sigUV1 = 1.0/sig2;
  double sigUV2 = 1.0/(sig2*K*K);
  double coeff1 = 1.0/(sig2*PI);
  double coeff2 = 1.0/(sig2*PI*K*K);
  
  Image *tmp_img = new Image(gsx,gsy);
  int i,j;
  for(i=-radii;i<=radii;i++){
    int v = i+radii; int i2 = i*i;
    for(j=-radii;j<=radii;j++){
      int u = j+radii;
      double dist = ((double)(i2+j*j));
      tmp_img->img[v][u] = coeff1*exp(-(dist*sigUV1))-coeff2*exp(-(dist*sigUV2));
      
    }
  }
  
  return tmp_img;
}

/* 畳み込み */
void Conv(Image *in,int radii,Image *kernel,Image *out){
  int i,j;
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      int i2,j2;
      out->img[i][j] = 0.0;
      for(i2=-radii;i2<=radii;i2++){
	int v = i2+radii;
	int v2 = i+i2;if(v2<0)v2=0;if(v2>in->sy-1)v2=in->sy-1;
	for(j2=-radii;j2<=radii;j2++){
	    int u = j2+radii;
	    int u2 = j+j2;if(u2<0)u2=0;if(u2>in->sx-1)u2=in->sx-1;
	    out->img[i][j] += in->img[v2][u2]*kernel->img[v][u];
	    
	  }
      }
      
    }
}

/* 正規化用の関数 */
void Normalize(Image *R,Image *G,Image *B){
  double max,min;
  max= min = R->img[0][0];
  int i,j;
  
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      if(max<R->img[i][j])max=R->img[i][j];
      if(max<G->img[i][j])max=G->img[i][j];
      if(max<B->img[i][j])max=B->img[i][j];
      if(min>R->img[i][j])min=R->img[i][j];
      if(min>G->img[i][j])min=G->img[i][j];
      if(min>B->img[i][j])min=B->img[i][j];
    }
  double range = fabs(max-min);if(range<=0.0)range=1.0;
   
  for(i=0;i<R->sy;i++)
     for(j=0;j<R->sx;j++){
       double val = 255.0*(R->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      R->img[i][j] = val;
      val = 255.0*(G->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      G->img[i][j] = val;
      val = 255.0*(B->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      B->img[i][j] = val;
     }
     
}
/* グレースケールの標準偏差を計算する関数 */
double getStandardDeviation(Image *R,Image *G,Image *B){
  double ave=0.0;
  double sd=0.0;
  int i,j;
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      double gray = (R->img[i][j]+G->img[i][j]+B->img[i][j])/3.0;
      ave += gray;
      sd += gray*gray;
    }
  ave /= ((double)(R->sx*R->sy));
  sd /= ((double)(R->sx*R->sy));
  sd -= ave*ave;
  sd = sqrt(sd);
  return sd;

}

/* Bilateralフィルタ */
void BilateralFilter(Image *R,Image *G,Image *B,Image *outR,Image *outG,Image *outB,int radii,double sigma_s,double sigma_I){
 

  /* 効率化のためGauss関数の中でsigSをかけるだけにするための変数 */
  double sigS = 1.0/(2.0*sigma_s*sigma_s);

  /* 輝度Gauss重みの標準偏差sigma_tへセットする(ただし今回は、sigma_t*(標準偏差)を使う) */
 
  /* 輝度値(グレースケール)の画像全体の標準偏差を計算し、standard_deviationへセットする */
  double standard_deviation = getStandardDeviation(R,G,B);
  printf("sigma_I = %lf, standard_deviation = %lf, (sigma_I*standard_deviation) = %lf\n",sigma_I,standard_deviation,(sigma_I*standard_deviation));
  /* 輝度Gauss重みの標準偏差sigma_tへ標準偏差をかける */
  double sigma_t =sigma_I*standard_deviation; 
  
  /* 効率化のためGauss関数の中でsigIをかけるだけにするための変数 */
  double sigI = 1.0/(2.0*sigma_t*sigma_t);
/* Bilateralフィルタここから */
  int sx = R->sx;
  int sy = R->sy;
  /* 各画素をたどるループ  */  
  int i,j;
  for(i=0;i<sy;i++)
    for(j=0;j<sx;j++){
      /* R,G,Bを足し合わせるための変数を用意し、ゼロを代入する */
      double dR=0.0;
      double dG=0.0;
      double dB=0.0;
      
      int l,m,u,v;
      
     
      double weight_sum=0.0;
      /* (2*radii+1)x(2*radii+1)の局所Window内をたどるループ */  
      for(l=-radii;l<=radii;l++){
	v = l+i;if(v<0)v=0;if(v>=sy)v=sy-1;
	for(m=-radii;m<=radii;m++){
	  u = m+j;if(u<0)u=0;if(u>=sx)u=sx-1;
	  
	  /* カラー勾配を計算 */
	  double difR = R->img[v][u]-R->img[i][j];
	  double difG = G->img[v][u]-G->img[i][j];
	  double difB = B->img[v][u]-B->img[i][j];
	  
	  /* カラー勾配強度(2乗)の計算 
	     注意点：3.0で割らなくてもよいが、グレースケールのときのパラメータと大きさを揃えるため3.0で割る
	   */
	  double grad = (difR*difR+difG*difG+difB*difB)/3.0;
	  /* 空間の距離 */
	  double dx = ((double)(m*m));
	  double dy = ((double)(l*l));
	  /* 重みの計算 G_{sigma_s}(x-u,y-v)*G_{sigma_t}(I(x,y)-I(u,v))*/
	  double weight = exp(-(grad*sigI+((dx+dy)*sigS)));
	  /* 正規化重み付き平均の為にweightの総和を計算 */
	  weight_sum+=weight;
	  /* 画素の座標[v][u]のR,G,B値にweightをかけdR,dG,dBへ足し合わせる:重み付き和  */
	  dR += R->img[v][u]*weight;
	  dG += G->img[v][u]*weight;
	  dB += B->img[v][u]*weight;
	  
	}
      }
      /* エラーチェック */
      if(weight_sum<=0.0)weight_sum=1.0;
      
      
      /* 出力のoutR,outG,outBへnormalized convolutionとして代入　*/
      outR->img[i][j] = dR/weight_sum;
      outG->img[i][j] = dG/weight_sum;
      outB->img[i][j] = dB/weight_sum;
      
      
    }
  
  
  /* Bilateralフィルタ計算ここまで */
  

}
/* Copy関数 */
void Copy(Image *R,Image *G,Image *B,Image *outR,Image *outG,Image *outB){
  int i,j;
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      outR->img[i][j] = R->img[i][j];
      outG->img[i][j] = G->img[i][j];
      outB->img[i][j] = B->img[i][j];
    }

}

/* 繰り返しBilateralフィルタ */
void IterativeBilateralFilter(int N,Image *R,Image *G,Image *B,Image *outR,Image *outG,Image *outB,int radii,double sigma_s,double sigma_I){
  Image *tmpR = new Image(R->sx,R->sy);
  Image *tmpG = new Image(R->sx,R->sy);
  Image *tmpB = new Image(R->sx,R->sy);
  int k;
  Copy(R,G,B,tmpR,tmpG,tmpB);
  for(k=0;k<N;k++){
    BilateralFilter(tmpR,tmpG,tmpB,outR,outG,outB,radii,sigma_s,sigma_I);
    if(k>=N-1)break;
    Copy(outR,outG,outB,tmpR,tmpG,tmpB);
  }
  delete tmpR;
  delete tmpG;
  delete tmpB;
}

/* 量子化 */
void Quantization(Image *in,Image *out,int binN){
  int i,j;
  double max=in->img[0][0];
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      if(max<in->img[i][j])max=in->img[i][j];
    }
  if(max<=0.0)max=1.0;
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      double val = ((double)(binN))*in->img[i][j]/max;
      int vali = ((int)(val));
      if(fabs(((double)(vali))-val)>=0.5)vali++;
      out->img[i][j] = ((double)(vali))*max/((double)(binN));
    }
}

/* maxで割る正規化 */
void Normalize(Image *in){
  int i,j;
  double max;
  max=in->img[0][0];
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      if(max<in->img[i][j])max=in->img[i][j];
    }
  max=fabs(max);
  if(max<=0.0)max=1.0;
   for(i=0;i<in->sy;i++)
     for(j=0;j<in->sx;j++)in->img[i][j]/=max;
   
}


/* HSV量子化: Hue色相だけbinNの数に量子化、alphaは明度Vの強調パラメータ */
void QuantizationHSV(Image *R,Image *G,Image *B,int binN,double alpha){
  int i,j;
  
  Normalize(R,G,B);
  Image *H = new Image(R->sx,R->sy);
  Image *S = new Image(R->sx,R->sy);
  Image *V = new Image(R->sx,R->sy);
  
  
  //RGB->HSV
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      double max,min;
      int checkmax=0;
      min = max = R->img[i][j];
      if(max<G->img[i][j]){
	max=G->img[i][j];
	checkmax=1;
      }
      if(max<B->img[i][j]){
	max=B->img[i][j];
	checkmax=2;
      }
      if(min>G->img[i][j])min=G->img[i][j];
      if(min>B->img[i][j])min=B->img[i][j];

      S->img[i][j] = (max-min);
      if(max>0.0)S->img[i][j]/=max;
      V->img[i][j] = max;
     
      double div=max-min;
      if(div<=0.0)div=1.0;
      if(checkmax==0){
	H->img[i][j] = 60.0*(G->img[i][j]-B->img[i][j])/div;
      }else if(checkmax==1){
	H->img[i][j] = 60.0*(B->img[i][j]-R->img[i][j])/div+120.0;
      }else{
	H->img[i][j] = 60.0*(R->img[i][j]-G->img[i][j])/div+240.0;
      }
      if(H->img[i][j]<0.0)H->img[i][j]+=360.0;
     
    }
  Normalize(S);
  Normalize(V);
  /* Vの強調 */
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      V->img[i][j] =  V->img[i][j]*20.0-10.0;
      V->img[i][j] = 255.0/(1.0+exp(-alpha*V->img[i][j]));
    }
  /* Hの量子化 */
  Quantization(H,H,binN);
  //HSV->RGB
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){

      if(S->img[i][j]==0.0){
	R->img[i][j]=G->img[i][j]=B->img[i][j]=V->img[i][j];
      }else{
      double Hi = H->img[i][j]/60.0;
      int Hii = ((int)(Hi));
      Hii = Hii % 6;
      
      double f = H->img[i][j]/60.0-((double)(Hii));
      double p = V->img[i][j]*(1.0-S->img[i][j]);
      double q = V->img[i][j]*(1.0-f*S->img[i][j]);
      double t = V->img[i][j]*(1.0-(1.0-f)*S->img[i][j]);
      if(Hii==0){
	R->img[i][j]=V->img[i][j];G->img[i][j]=t;B->img[i][j]=p;
      }else if(Hii==1){
	R->img[i][j]=q;G->img[i][j]=V->img[i][j];B->img[i][j]=p;
      }else if(Hii==2){
	R->img[i][j]=p;G->img[i][j]=V->img[i][j];B->img[i][j]=t;
      }else if(Hii==3){
	R->img[i][j]=p;G->img[i][j]=q;B->img[i][j]=V->img[i][j];
      }else if(Hii==4){
	R->img[i][j]=t;G->img[i][j]=p;B->img[i][j]=V->img[i][j];
      }else{
	R->img[i][j]=V->img[i][j];G->img[i][j]=p;B->img[i][j]=q;
      }
      }
      if(R->img[i][j]<0.0)R->img[i][j]=0.0;
      if(G->img[i][j]<0.0)G->img[i][j]=0.0;
      if(B->img[i][j]<0.0)B->img[i][j]=0.0;
      
      if(R->img[i][j]>255.0)R->img[i][j]=255.0;
      if(G->img[i][j]>255.0)G->img[i][j]=255.0;
      if(B->img[i][j]>255.0)B->img[i][j]=255.0;
      
    }
  delete H;
  delete S;
  delete V;

}



/* g++ -o Style Style.cxx -lm
引数11個
./Style 入力ファイル名 出力ファイル名 畳み込みの半径(int) DoGの空間標準偏差(double) DoGのバンド幅(double) Bilateralフィルタの空間標準偏差(double) Bilateralフィルタの輝度標準偏差(double) Bilateralフィルタの繰り返し回数(int)　HSV量子化数(int) HSV量子化V強調パラメータ(double) RGB量子化数(int) 

例えば、
./Style lena.bmp style_lena.bmp 10 0.5 3.0 25.0 0.1 3 16 0.5 4

*/

int main(int argc,char *argv[]){
  int i,j;
  int sx,sy;
  /* BMP画像の読み込み */
  BMPIO *myio = new BMPIO();
  myio->readBMPSize(&sx,&sy,argv[1]);
  Image *R = new Image(sx,sy);
  Image *G = new Image(sx,sy);
  Image *B = new Image(sx,sy);
  myio->readBMP(R,G,B,argv[1]);
  

  /* 計算の為の画像 */
  Image *gray = new Image(sx,sy);
  Image *out_gray = new Image(sx,sy);
  Image *outR = new Image(sx,sy);
  Image *outG = new Image(sx,sy);
  Image *outB = new Image(sx,sy);
  
  /* 畳み込みの半径 */
  int radii = atoi(argv[3]);
  /* DoGの空間標準偏差(double) */
  double sig = atof(argv[4]);
  /* DoGのバンド幅(double) */
  double K = atof(argv[5]);
  /* Bilateralフィルタの空間標準偏差(double) */
  double sigBS = atof(argv[6]);
  /* Bilateralフィルタの輝度標準偏差(double) */
  double sigBI = atof(argv[7]);
  /* Bilateralフィルタの繰り返し回数(int) */
  int N = atoi(argv[8]);
  /* HSVのHを多値化する数 */
  int HSV_N = atoi(argv[9]);
  /* HSV量子化のV強調パラメータ */
  double alpha = atof(argv[10]);
  /* RGBを多値化する数 */
  int RGB_N = atoi(argv[11]);

  
  /* Bilateralフィルタの繰り返し */ 
  IterativeBilateralFilter(N,R,G,B,outR,outG,outB,radii,sigBS,sigBI);
  /* outR,outG,outBを0-255へ線形補間で正規化する */
  Normalize(outR,outG,outB);
  /* グレースケール画像の作成 */
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++)
      gray->img[i][j] = (outR->img[i][j]+outG->img[i][j]+outB->img[i][j])/3.0;
  
  /* HSV空間での多値化(量子化) */
  QuantizationHSV(outR,outG,outB,HSV_N,alpha);
  
  /* RGB空間での多値化(量子化)　*/
  Quantization(outR,outR,RGB_N);
  Quantization(outG,outG,RGB_N);
  Quantization(outB,outB,RGB_N);
    
  
  
  
  
  /* DoG画像の生成 */
  Image *dog = makeDoGImage(radii,sig,K);
  /* DoGの計算 */
  Conv(gray,radii,dog,out_gray);
  
  /* Stylized画像の合成 */
  for(i=0;i<out_gray->sy;i++)
    for(j=0;j<out_gray->sx;j++){
        if(out_gray->img[i][j]>0.0){
	  /* 繰り返しBilateralフィルタ->HSV多値化->RGB多値化　*/
	outR->img[i][j] = outR->img[i][j];
	outG->img[i][j] = outG->img[i][j];
	outB->img[i][j] = outB->img[i][j];
	
      }else{
	
	  /* 繰り返しBilateralフィルタ->DoGによるEdge */
	outR->img[i][j] = 0.0;
	outG->img[i][j] = 0.0;
	outB->img[i][j] = 0.0;
	
      }
      
      
    }
  
  /* 出力Stylized画像の保存 */
  myio->saveBMP(outR,outG,outB,argv[2]);
  
  /* メモリの開放 */
  delete dog;
  delete gray;
  delete out_gray;

  delete R;
  delete G;
  delete B;

  delete outR;
  delete outG;
  delete outB;

  delete myio;

  return 0;
}

#undef PI
